#!/usr/bin/env python
import sys
import time
import random

import RPi.GPIO as GPIO
GPIO.setmode(GPIO.BCM)
GPIO.setwarnings(False)

pinchDelay = .003
delay = .002 #delay between pulses for max speed on single windings
## define sets of motor coil gpio control pins
A1   = 15 #pin 10
A2   = 14 #pin 8
B1   = 18 #pin 12
B2   = 23 #pin 16
FBA1 = 9  #pin 23
FBA2 = 11 #pin 21
FBB1 = 10 #pin 19
FBB2 = 22 #pin 15
LRA1 = 17 #pin 13
LRA2 = 27 #pin 11
LRB1 = 4  #pin 7
LRB2 = 3  #pin 5


## define enable pins for L293 contollers. These GPIO turn the motor controllers
## on and of
Back =  2  #pin 3  (yellow)
Left =  8  #pin 24 (green)
Front = 25 #pin 22 (white)
Right = 24 #pin 18 (blue)

Interrupt = 7 #pin 26

GPIO.setup(A1,GPIO.OUT)
GPIO.setup(A2,GPIO.OUT)
GPIO.setup(B1,GPIO.OUT)
GPIO.setup(B2,GPIO.OUT)
GPIO.setup(FBA1,GPIO.OUT)
GPIO.setup(FBA2,GPIO.OUT)
GPIO.setup(FBB1,GPIO.OUT)
GPIO.setup(FBB2,GPIO.OUT)
GPIO.setup(LRA1,GPIO.OUT)
GPIO.setup(LRA2,GPIO.OUT)
GPIO.setup(LRB1,GPIO.OUT)
GPIO.setup(LRB2,GPIO.OUT)
GPIO.setup(Front,GPIO.OUT)
GPIO.setup(Back,GPIO.OUT)
GPIO.setup(Right,GPIO.OUT)
GPIO.setup(Left ,GPIO.OUT)
##GPIO.setup(Interrupt, GPIO.IN)

rotateSequence = [[1,0,0,1],[0,0,0,1],[0,1,0,1],[0,1,0,0],[0,1,1,0],[0,0,1,0],[1,0,1,0],[1,0,0,0]]  #rotate sequence
DirCount = [0,0,0,0] #[N,S,E,W] #tracks stepper counts
globalState = 0  ## State to represent orientation of cube
globalDirection = 1  ## Flag for move direction; 1 is clockwise, 0 counter-clockwise
globalLRPosition = 'closed'
globalFBPosition = 'closed'

## List of states
stateList = list()

## State 0: Red up, Green right
stateList.append({"R": "U", "O": "D", "G": "L", "B": "R", "W": "F", "Y": "B"})

## State 1: Red right, Blue up
stateList.append({"R": "L", "O": "R", "B": "U", "G": "D", "W": "F", "Y": "B"})

## State 2: Red right, White up
stateList.append({"R": "L", "O": "R", "B": "B", "G": "F", "W": "U", "Y": "D"})

## State 3: Red up, Yellow right
stateList.append({"R": "U", "O": "D", "B": "B", "G": "F", "W": "R", "Y": "L"})

## State 4: Red front, Blue up
stateList.append({"R": "F", "O": "B", "B": "U", "G": "D", "W": "R", "Y": "L"})

## State 5: Red front, Yellow up
stateList.append({"R": "F", "O": "B", "B": "R", "G": "L", "W": "D", "Y": "U"})


## rotate a single side 90 degrees
## motor: names are B-Back F-Front R-Right L-Left
## way:   'clockwise' for clockwise, 'counter' for counter-clockwise, '180' for 180
## count: tracks motor position so arm always stops at correct steps
def RoSide(motor,way):

  global DirCount
  comp = 8        ## variable used to compensate for tolerance in claw
  
  if(way == '180'):
    steps = 200
    dir = 1  ## step 1 by 1, even if 180
  else:
    if(way == 'clockwise'):
      dir = 1
    else:
      dir = -1
    steps = 100

  ## enable motor control and set count
  if(motor == 'B'):
    GPIO.output(Back,True)
    count = DirCount[0]

  if(motor == 'F'):
    GPIO.output(Front,True)
    count = DirCount[1]
    dir = dir * -1

  if(motor == 'R'):
    GPIO.output(Right,True)
    count = DirCount[2]
    dir = dir * -1

  if(motor == 'L'):
    GPIO.output(Left,True)
    count = DirCount[3]


  ## perform the turns
  for i in range(0, steps + comp):
    setStep(rotateSequence[count])
    count = (count+dir)%8
    time.sleep(delay)
  for i in range(0, comp):
    count = (count-dir)%8
    setStep(rotateSequence[count])
    time.sleep(delay)

  ## turn motor off and update count
  ## DirCount is 0 for vertical, non-zero when horizontal
  if(motor == 'B'):
    GPIO.output(Back,False)
    DirCount[0] = count
  if(motor == 'F'):
    GPIO.output(Front,False)
    DirCount[1] = count
  if(motor == 'R'):
    GPIO.output(Right,False)
    DirCount[2] = count
  if(motor == 'L'):
    GPIO.output(Left,False)
    DirCount[3] = count

    

## Rotate entire cube 90 degrees
## motor names FB-Front/Back, LR-Left,Right
## all other funtions are same as RoSide()
def RoCube(motor,dir):
  comp = 4
  global DirCount
  ## enable motor control and set count
  if(motor == 'FB'):
    GPIO.output(Back,True)
    GPIO.output(Front,True)
    count = DirCount[0]

  elif(motor == 'LR'):
    GPIO.output(Right,True)
    GPIO.output(Left,True)
    count = DirCount[2]

  if(dir == 2):
    steps = 200
    dir = 1
  else:
    steps = 100

  for i in range(0, steps + comp):
    setStep(rotateSequence[count])
    count = (count+dir)%8
    time.sleep(delay)

  for i in range(0, comp):
    count = (count-dir)%8
    setStep(rotateSequence[count])
    time.sleep(delay)

  ## turn motor off and update count
  if(motor == 'FB'):
    GPIO.output(Back,False)
    GPIO.output(Front,False)
    DirCount[0] = count
    DirCount[1] = count
  if(motor == 'LR'):
    GPIO.output(Right,False)
    GPIO.output(Left,False)
    DirCount[2] = count
    DirCount[3] = count


## Tighen the arms of the robot before starting the solve sequence
def tighten():
  for i in range (0, 24):
    setStepFB(rotateSequence[i%8])
    time.sleep(pinchDelay)
  for i in range (0, 24):
    setStepLR(rotateSequence[i%8])
    time.sleep(pinchDelay)
  setStep(rotateSequence[0])
  GPIO.output(Back, True)
  GPIO.output(Front, True)
  GPIO.output(Right, True)
  GPIO.output(Left, True)
  time.sleep(delay)
  GPIO.output(Back, False)
  GPIO.output(Front, False)
  GPIO.output(Right, False)
  GPIO.output(Left, False)

## Controls pinching arms
## motor - names FB-Front/Back, LR-Left/Right
## way -   'open or 'close
## count - keeps track of pinch sequence
def Pinch(motor,way):

  count = 0
  if(way == 'open'):
    steps = 88
    dir = -1
  else:
    steps = 96
    dir = 1
     
  if(motor == 'FB'):
    for i in range(0,steps):
      setStepFB(rotateSequence[count%8])
      count += dir
      time.sleep(pinchDelay)

  if(motor =='LR'):
    for i in range(0,steps):
      setStepLR(rotateSequence[count%8])
      count += dir
      time.sleep(pinchDelay)

##Helper method for rotating motor
##controls Back and Left motors
def setStep(w):
  GPIO.output(A1,w[0])
  GPIO.output(A2,w[1])
  GPIO.output(B1,w[2])
  GPIO.output(B2,w[3])

def setStepFB(w):
  GPIO.output(FBA1,w[0])
  GPIO.output(FBA2,w[1])
  GPIO.output(FBB1,w[2])
  GPIO.output(FBB2,w[3])

def setStepLR(w):
  GPIO.output(LRA1,w[0])
  GPIO.output(LRA2,w[1])
  GPIO.output(LRB1,w[2])
  GPIO.output(LRB2,w[3])

##determines string for executring moves
##The string will start with an int to indicate its current state. This int is
##  removed when passing to solve(str)
def findRotations(moveSolution):
  ##print(moveSolution)
  moveSolution = moveSolution.split()
  moveList = list()
  moveList.append('0')    ##initalize move list with zero indicating state 0
  trimLength = 4*len(moveSolution)
  for getMove in moveSolution:  ##solve one move at a time
    moveSuffix = ''
    if len(getMove) == 2 :
      moveSuffix = getMove[1]  ##place holder for suffix
    for i in range(len(moveList)):
      s = moveList.pop(0)      ##get next branch in tree
      if(len(s)<trimLength):
        currentState = ord(s[0]) - ord('0')
        singleMove = stateList[currentState].get(getMove[0])
        if((singleMove == 'U') | (singleMove == 'D')):
          if(currentState%2):   ##if the state is even then -1 = LR; +1 = FB
            toAppend = chr(ord('0')+(currentState+1)%6) + s[1:len(s)] + " LR "
            toAppend = toAppend + stateList[(currentState+1)%6].get(getMove[0]) + moveSuffix  
            moveList.append(toAppend)
            toAppend = chr(ord('0')+(currentState-1)%6) + s[1:len(s)] + " FB "
            toAppend = toAppend + stateList[(currentState-1)%6].get(getMove[0]) + moveSuffix
            moveList.append(toAppend)
          else:
            toAppend = chr(ord('0')+(currentState+1)%6) + s[1:len(s)] + " FB "
            toAppend = toAppend + stateList[(currentState+1)%6].get(getMove[0]) + moveSuffix
            moveList.append(toAppend)
            toAppend = chr(ord('0')+(currentState-1)%6) + s[1:len(s)] + " LR "
            toAppend = toAppend + stateList[(currentState-1)%6].get(getMove[0]) + moveSuffix
            moveList.append(toAppend)
        else:
          toAppend = s + " " + singleMove + moveSuffix
          moveList.append(toAppend)
##  print moveList
  leastMovesLength = len(moveList[0])
  leastMovesPosition = 0
  for i in range(1,len(moveList)):
    if(leastMovesLength > len(moveList[i])):
       leastMovesLength = len(moveList[i])
       leastMovesPosition = i
  ##print(moveList[leastMovesPosition][2:])
  solve(moveList[leastMovesPosition][2:])
       
## Takes each move from solution string and performs rotations
def solve(moves):
  global globalDirection
  moves = moves.split()
  
  ## Iterates through the sequence of moves
  for singleMove in moves:
    globalDirection = 1
    if len(singleMove) == 2:
      if singleMove[1] == "'":
        globalDirection = 0	        ##set as CCW
        execute(singleMove[0])
      elif(singleMove[1] == '2'):
        globalDirection = 2		##set as 180 degree turn
        execute(singleMove[0])
      else:
        execute(singleMove)
    else:
      execute(singleMove)
    ##raw_input("...")
  setStepLR([0,0,0,0])
  setStepFB([0,0,0,0])

## Executes the appropriate move
def execute(move):
  time.sleep(delay)
  if move == "FB":
    frontBack()
  elif move == "LR":
    leftRight()
  elif move == "L":
    left()
  elif move == "R":
    right()
  elif move == "F":
    front()
  elif move == "B": 
    back()
  else:
    print(move)
    print("Invalid Move")

## rotates the cube so that the top is now on the right side
def frontBack():
  global globalState
  global globalFBPosition
  global globalLRPosition
  if(globalFBPosition == 'open'):
    Pinch('FB', 'closed')
  if(globalLRPosition == 'closed'):
    Pinch('LR', 'open')
  if ((globalState == 0) | (globalState == 3) | (globalState == 5)):
    RoCube('FB', -1)
    Pinch('LR', 'close')
    Pinch('FB', 'open')
    RoCube('FB', 1)
  if ((globalState == 1) | (globalState == 2) | (globalState == 4)):
    RoCube('FB', 1)
    Pinch('LR', 'close')
    Pinch('FB', 'open')
    RoCube('FB', -1)
  
  ## Change cube state after cube rotation
  globalFBPosition = 'open'
  globalLRPosition = 'closed'
  if(globalState%2):
    globalState = (globalState - 1) % 6
  else:
    globalState = (globalState + 1) % 6
     
## rotates cube so that the bottom is now on the left side
def leftRight():
  global globalState
  global globalLRPosition
  global globalFBPosition
  if(globalLRPosition == 'open'):
    Pinch('LR', 'closed')
  if(globalFBPosition == 'closed'):
    Pinch('FB', 'open')  
  if ((globalState == 0) | (globalState == 2) | (globalState == 3)):
    RoCube('LR', -1)
    Pinch('FB', 'close')
    Pinch('LR', 'open')
    RoCube('LR', 1)
  if ((globalState == 1) | (globalState == 4) | (globalState == 5)):
    RoCube('LR', 1)
    Pinch('FB', 'close')
    Pinch('LR', 'open')
    RoCube('LR', -1)

  ## Change cube state after cube rotation
  globalLRPosition = 'open'
  globalFBPosition = 'closed'
  if(globalState%2):
    globalState = (globalState + 1) % 6
  else:
    globalState = (globalState - 1) % 6

## rotates left face of cube
def left():
  global globalLRPosition
  global globalFBPosition
  if(globalLRPosition == 'open'):
    Pinch('LR', 'closed')
  if(globalFBPosition == 'open'):
    Pinch('FB', 'closed')
  if globalDirection == 2:
    RoSide('L', '180')
    globalFBPosition = 'closed'
    globalLRPosition = 'closed'

  elif globalDirection == 0:
    RoSide('L', 'counter')
    Pinch('LR', 'open')
    RoSide('L', 'clockwise')
    globalFBPosition = 'closed'
    globalLRPosition = 'open'

  else:
    RoSide('L', 'clockwise')
    Pinch('LR', 'open')
    RoSide('L', 'counter')
    globalFBPosition = 'closed'
    globalLRPosition = 'open'

## rotates right face of cube
def right():
  global globalLRPosition
  global globalFBPosition
  if(globalLRPosition == 'open'):
    Pinch('LR', 'closed')
  if(globalFBPosition == 'open'):
    Pinch('FB', 'closed')
    
  if globalDirection == 2:
    RoSide('R', '180')
    globalFBPosition = 'closed'
    globalLRPosition = 'closed'

  elif globalDirection == 0:
    RoSide('R', 'counter')
    Pinch('LR', 'open')
    RoSide('R', 'clockwise')
    globalFBPosition = 'closed'
    globalLRPosition = 'open'

  else:
    RoSide('R', 'clockwise')
    Pinch('LR', 'open')
    RoSide('R', 'counter')
    globalFBPosition = 'closed'
    globalLRPosition = 'open'

## rotates front face of cube
def front():
  global globalLRPosition
  global globalFBPosition
  if(globalLRPosition == 'open'):
    Pinch('LR', 'closed')
  if(globalFBPosition == 'open'):
    Pinch('FB', 'closed')
    
  if globalDirection == 2:
    RoSide('F', '180')
    globalFBPosition = 'closed'
    globalLRPosition = 'closed'
    
  elif globalDirection == 0:
    RoSide('F', 'counter')
    Pinch('FB', 'open')
    RoSide('F', 'clockwise')
    globalFBPosition = 'open'
    globalLRPosition = 'closed'

  else:
    RoSide('F', 'clockwise')
    Pinch('FB', 'open')
    RoSide('F', 'counter')
    globalFBPosition = 'open'
    globalLRPosition = 'closed'
    
## rotates rear face of cube
def back():
  global globalLRPosition
  global globalFBPosition
  if(globalLRPosition == 'open'):
    Pinch('LR', 'closed')
  if(globalFBPosition == 'open'):
    Pinch('FB', 'closed')
    
  if globalDirection == 2:
    RoSide('B', '180')
    globalFBPosition = 'closed'
    globalLRPosition = 'closed'
    
  elif globalDirection == 0:
    RoSide('B', 'counter')
    Pinch('FB', 'open')
    RoSide('B', 'clockwise')
    globalFBPosition = 'open'
    globalLRPosition = 'closed'
    
  else:
    RoSide('B', 'clockwise')
    Pinch('FB', 'open')
    RoSide('B', 'counter')
    globalFBPosition = 'open'
    globalLRPosition = 'closed'


## Rotates cube for image capturing, incorporates communication with iOS
def autoRotate(count):
  global DirCount

  ## move hold red side with arms out of the way
  if(count == '0'):
    tighten()
    Pinch('LR', 'open')
    RoCube('LR', 1)
    Pinch('LR', 'close')
    Pinch('FB', 'open')
    print("Arms Cleared")

  ## rotate to white side
  elif(count == '1'):
    DirCount[2] = 4
    RoCube('LR', 1)
    RoCube('FB', 1)
    Pinch('FB', 'close')
    Pinch('LR', 'open')
    print("Rotate Finished")

  ## rotate to green side
  elif(count == '2'):
    DirCount[0] = 4
    RoCube('FB', 1)
    Pinch('LR', 'close')
    Pinch('FB', 'open')
    RoCube('FB', 1)
    Pinch('FB', 'close')
    Pinch('LR', 'open')
    print("Rotate Finished")

  ## rotate to yellow side
  elif(count == '3'):
    DirCount[0] = 4
    RoCube('FB', 1)
    Pinch('LR', 'close')
    Pinch('FB', 'open')
    RoCube('FB', 1)
    Pinch('FB', 'close')
    Pinch('LR', 'open')
    print("Rotate Finished")

  ## rotate to blue side
  elif(count == '4'):
    DirCount[0] = 4
    RoCube('FB', 1)
    Pinch('LR', 'close')
    Pinch('FB', 'open')
    RoCube('FB', 1)
    Pinch('FB', 'close')
    Pinch('LR', 'open')
    print("Rotate Finished")

  ## rotate to orange side
  elif(count == '5'):
    DirCount[0] = 4
    RoCube('FB', 1)
    Pinch('LR', 'close')
    Pinch('FB', 'open')
    RoCube('LR', 1)
    print("Rotate Finished")

  ## rotate to red, reset arms to ready position
  elif(count == '6'):
    DirCount[2] = 4
    RoCube('LR', 2)
    Pinch('FB', 'close')
    Pinch('LR', 'open')
    RoCube('LR', 1)
    Pinch('LR', 'close')
    print("Rotate Finished")

def allOff(): 
  GPIO.output(Back, False)
  GPIO.output(Front, False)
  GPIO.output(Right, False)
  GPIO.output(Left, False)
  GPIO.output(FBA1, False)
  GPIO.output(FBA2, False)
  GPIO.output(FBB1, False)
  GPIO.output(FBB2, False)
  GPIO.output(LRA1, False)
  GPIO.output(LRA2, False)
  GPIO.output(LRB1, False)
  GPIO.output(LRB2, False)

## Run solving script on command-line argument.
## Moves in solution string must be separated by single space

autoRotate(sys.argv[1])

## count 7 means ready to solve/perform moves
if(sys.argv[1] == '7'):
  findRotations(sys.argv[2])
  allOff()

if(sys.argv[1] == 'zero'):	## used to center the claws if they rotated
  setStep(rotateSequence[0])
  GPIO.output(Back, True)
  raw_input("align yellow")
  GPIO.output(Back, False)
  GPIO.output(Front, True)
  raw_input("align white")
  GPIO.output(Front, False)
  GPIO.output(Right, True)
  raw_input("align Blue")
  GPIO.output(Right, False)
  GPIO.output(Left, True)
  raw_input("align Green")
  GPIO.output(Left, False)
    
elif(sys.argv[1] == 'test'):  ## several general test that can be ran
  tighten()
  if(sys.argv[2] == '1'):	## several moves requiring cube rotations
    findRotations("G' Y2 B R B2 R2 W' G") 
  elif(sys.argv[2] == '2'):	## opposite of test one
    findRotations("G' W R2  B2 R' B' Y2 G")
  elif(sys.argv[2] == '3'):
    findRotations("G Y B W")
  elif(sys.argv[2] == '5'):
    for i in range(0,15):
      leftRight()
      front()
  elif(sys.argv[2] == '6'):
    for i in range(0,15):
      frontBack()
      right()
  elif(sys.argv[2] == '7'):  
    while(1):    
      findRotations("G' Y2 B R B2 R2 W' G") 
      Pinch('LR','close')
      time.sleep(5)
      findRotations("G' W R2  B2 R' B' Y2 G")
      Pinch('LR','close')
      time.sleep(5)
  allOff()


